#define ARDUINO_ESP8266_GIT_VER   0xcf6ff4c4
#define ARDUINO_ESP8266_GIT_DESC  3.0.2
#define ARDUINO_ESP8266_VERSION   3.0.2

#define ARDUINO_ESP8266_MAJOR     3
#define ARDUINO_ESP8266_MINOR     0
#define ARDUINO_ESP8266_REVISION  2

#define ARDUINO_ESP8266_RELEASE   "3.0.2"
#define ARDUINO_ESP8266_RELEASE_3_0_2
