#ifndef ESP_SDK_VERSION_H
#define ESP_SDK_VERSION_H

#define ESP_SDK_VERSION_MAJOR   2
#define ESP_SDK_VERSION_MINOR   2
#define ESP_SDK_VERSION_PATCH   0

#define ESP_SDK_VERSION_NUMBER 0x020200
#define ESP_SDK_VERSION_STRING  "2.2.0"

#endif
