// pgmspace.h stub

// This file's contents have been moved to newlib.  This file simply
// includes the newlib pgmspace file as well as some ets headers
// to preserve backwards compatibility
// current source: https://github.com/earlephilhower/newlib-xtensa/blob/xtensa-4_0_0-lock-arduino/newlib/libc/sys/xtensa/sys/pgmspace.h

#include <sys/pgmspace.h>

#ifdef __ets__

#include "ets_sys.h"
#include "osapi.h"

#endif
