
/*
  pins_arduino.h - Pin definition functions for Arduino
  Part of Arduino - http://www.arduino.cc/
  Copyright (c) 2007 David A. Mellis
  Modified for ESP8266 platform by Ivan Grokhotkov, 2014-2015.
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  You should have received a copy of the GNU Lesser General
  Public License along with this library; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA  02111-1307  USA
  $Id: wiring.h 249 2007-02-03 16:52:51Z mellis $
*/

#ifndef Pins_Arduino_h
#define Pins_Arduino_h

#include "../generic/common.h"

#ifndef ESPECTRO_CORE_VERSION
#define ESPECTRO_CORE_VERSION   3
#endif

static const uint8_t SDA = 4;
static const uint8_t SCL = 5;

#define LED_BUILTIN 15

#if ESPECTRO_CORE_VERSION == 3
static const uint8_t BUTTON_BUILTIN = 0;
static const uint8_t BUILTIN_BUTTON = 0;
#else
static const uint8_t BUTTON_BUILTIN = 2;
static const uint8_t BUILTIN_BUTTON = 2;
#endif

static const uint8_t RX   	= 3;
static const uint8_t TX   	= 1;
static const uint8_t RX0   	= 3;
static const uint8_t TX0   	= 1;
static const uint8_t TX1   	= 2;

#endif /* Pins_Arduino_h */