#ifndef MegaGpioPinMap_h
#define MegaGpioPinMap_h


namespace sdfat {


static const GpioPinMap_t GpioPinMap[] = {
  GPIO_PIN(E, 0),  // D0
  GPIO_PIN(E, 1),  // D1
  GPIO_PIN(E, 4),  // D2
  GPIO_PIN(E, 5),  // D3
  GPIO_PIN(G, 5),  // D4
  GPIO_PIN(E, 3),  // D5
  GPIO_PIN(H, 3),  // D6
  GPIO_PIN(H, 4),  // D7
  GPIO_PIN(H, 5),  // D8
  GPIO_PIN(H, 6),  // D9
  GPIO_PIN(B, 4),  // D10
  GPIO_PIN(B, 5),  // D11
  GPIO_PIN(B, 6),  // D12
  GPIO_PIN(B, 7),  // D13
  GPIO_PIN(J, 1),  // D14
  GPIO_PIN(J, 0),  // D15
  GPIO_PIN(H, 1),  // D16
  GPIO_PIN(H, 0),  // D17
  GPIO_PIN(D, 3),  // D18
  GPIO_PIN(D, 2),  // D19
  GPIO_PIN(D, 1),  // D20
  GPIO_PIN(D, 0),  // D21
  GPIO_PIN(A, 0),  // D22
  GPIO_PIN(A, 1),  // D23
  GPIO_PIN(A, 2),  // D24
  GPIO_PIN(A, 3),  // D25
  GPIO_PIN(A, 4),  // D26
  GPIO_PIN(A, 5),  // D27
  GPIO_PIN(A, 6),  // D28
  GPIO_PIN(A, 7),  // D29
  GPIO_PIN(C, 7),  // D30
  GPIO_PIN(C, 6),  // D31
  GPIO_PIN(C, 5),  // D32
  GPIO_PIN(C, 4),  // D33
  GPIO_PIN(C, 3),  // D34
  GPIO_PIN(C, 2),  // D35
  GPIO_PIN(C, 1),  // D36
  GPIO_PIN(C, 0),  // D37
  GPIO_PIN(D, 7),  // D38
  GPIO_PIN(G, 2),  // D39
  GPIO_PIN(G, 1),  // D40
  GPIO_PIN(G, 0),  // D41
  GPIO_PIN(L, 7),  // D42
  GPIO_PIN(L, 6),  // D43
  GPIO_PIN(L, 5),  // D44
  GPIO_PIN(L, 4),  // D45
  GPIO_PIN(L, 3),  // D46
  GPIO_PIN(L, 2),  // D47
  GPIO_PIN(L, 1),  // D48
  GPIO_PIN(L, 0),  // D49
  GPIO_PIN(B, 3),  // D50
  GPIO_PIN(B, 2),  // D51
  GPIO_PIN(B, 1),  // D52
  GPIO_PIN(B, 0),  // D53
  GPIO_PIN(F, 0),  // D54
  GPIO_PIN(F, 1),  // D55
  GPIO_PIN(F, 2),  // D56
  GPIO_PIN(F, 3),  // D57
  GPIO_PIN(F, 4),  // D58
  GPIO_PIN(F, 5),  // D59
  GPIO_PIN(F, 6),  // D60
  GPIO_PIN(F, 7),  // D61
  GPIO_PIN(K, 0),  // D62
  GPIO_PIN(K, 1),  // D63
  GPIO_PIN(K, 2),  // D64
  GPIO_PIN(K, 3),  // D65
  GPIO_PIN(K, 4),  // D66
  GPIO_PIN(K, 5),  // D67
  GPIO_PIN(K, 6),  // D68
  GPIO_PIN(K, 7)   // D69
};


}; // namespace sdfat


#endif  // MegaGpioPinMap_h
